// 
// このコードは、https://demo.mescius.jp/diodocs/pdf のサンプルの一部として提供されています。
// © MESCIUS inc. All rights reserved.
// 
using System;
using System.IO;
using System.Drawing;
using System.Numerics;
using System.Linq;
using System.Collections.Generic;
using GrapeCity.Documents.Pdf;
using GrapeCity.Documents.Text;
using GrapeCity.Documents.Drawing;
using GrapeCity.Documents.Svg;
using GCTEXT = GrapeCity.Documents.Text;
using GCDRAW = GrapeCity.Documents.Drawing;
using DsPdfWeb.Demos.Common;

namespace DsPdfWeb.Demos
{
    // このサンプルでは、GcSvgDocument と GcGraphics.DrawSvg() を使用して、いくつかの幾何学図形を描画します。
    // サンプルのコードと結果は、Shapes サンプルと同じようなものになります。
    public class SvgShapes
    {
        // 多角形とその下にキャプションを描画するヘルパーメソッドです。
        // 実際の描画を行わずに位置を計算するためにも使用できます。
        // startAngle は（1,0）から時計回りの最初の点です。
        // Points は「中心」を基準とした相対的なものです。
        private SvgPoint[] DrawPolygon(GcGraphics g, PointF center, float r, int n, float startAngle, SvgPaint stroke, string caption = null)
        {
            const float Q = 1.4f;
            var plgn = new SvgPolygonElement();
            plgn.Points = new List<SvgPoint>(n);
            for (int i = 0; i < n; ++i)
                plgn.Points.Add(new SvgPoint(
                    new SvgLength((float)(r * Math.Cos(startAngle + 2 * Math.PI * i / n)) * Q, SvgLengthUnits.Percentage),
                    new SvgLength((float)(r * Math.Sin(startAngle + 2 * Math.PI * i / n)) * Q, SvgLengthUnits.Percentage)));

            if (stroke != null)
            {
                plgn.Stroke = stroke;
                plgn.Fill = new SvgPaint(SvgColor.Transparent);
                var svgDoc = new GcSvgDocument();
                var svg = svgDoc.RootSvg;
                svg.Children.Add(plgn);
                g.DrawSvg(svgDoc, new RectangleF(center.X, center.Y, r * 2, r * 2));
            }

            if (!string.IsNullOrEmpty(caption))
                DrawCaption(g, center, r, caption);

            return plgn.Points.ToArray();
        }
        // 図形の下にキャプションを描画するヘルパーメソッドです。
        private void DrawCaption(GcGraphics g, PointF center, float r, string caption)
        {
            g.DrawString(caption,
                new TextFormat() { Font = Util.getFont(), FontSize = 10, },
                new RectangleF(center.X - r, center.Y + r, r * 2, 24),
                TextAlignment.Center, ParagraphAlignment.Center, false);
        }
        // メインエントリポイントです。
        public int CreatePDF(Stream stream)
        {
            var doc = new GcPdfDocument();
            var page = doc.Pages.Add();
            var g = page.Graphics;
            // ドキュメントのヘッダーです。
            g.DrawString("図形 (SVG)",
                new TextFormat() { Font = Util.getFont(), FontSize = 14, Underline = true, },
                new RectangleF(PointF.Empty, new SizeF(page.Size.Width, 44)),
                TextAlignment.Center, ParagraphAlignment.Far);

            // レイアウト用のグリッドを設定します。
            var grid = new
            {
                Cols = 3,
                Rows = 5,
                MarginX = 72,
                MarginY = 36,
                Radius = 36,
                StepX = (page.Size.Width - 144) / 3,
                StepY = (page.Size.Height - 72) / 5,
            };

            // 次の図の中心の挿入位置です。
            var startIp = new PointF(grid.MarginX + grid.StepX / 2, grid.MarginY + grid.StepY / 2 + 10);
            var ip = startIp;
            // レイアウトグリッドを表示するためのデバッグコードです。
            /*
            var ipp = ip;
            for (int i = 0; i < grid.Cols; ++i)
            {
                ipp.Y = ip.Y;
                for (int j = 0; j < grid.Rows; ++j)
                {
                    g.DrawRectangle(new RectangleF(ipp.X - grid.Radius, ipp.Y - grid.Radius, grid.Radius * 2, grid.Radius * 2), Color.LightGreen, 0.5f);
                    ipp.Y += grid.StepY;
                }
                ipp.X += grid.StepX;
            }
            */

            // 再利用可能な SVG の基本要素です。
            var len0 = new SvgLength(0, SvgLengthUnits.Percentage);
            var len10 = new SvgLength(10, SvgLengthUnits.Percentage);
            var len20 = new SvgLength(20, SvgLengthUnits.Percentage);
            var len30 = new SvgLength(30, SvgLengthUnits.Percentage);
            var len40 = new SvgLength(40, SvgLengthUnits.Percentage);
            var len50 = new SvgLength(50, SvgLengthUnits.Percentage);
            var len60 = new SvgLength(60, SvgLengthUnits.Percentage);
            var len70 = new SvgLength(70, SvgLengthUnits.Percentage);
            var len80 = new SvgLength(80, SvgLengthUnits.Percentage);
            var len90 = new SvgLength(90, SvgLengthUnits.Percentage);
            var len100 = new SvgLength(100, SvgLengthUnits.Percentage);
            var paintTrans = new SvgPaint(SvgColor.Transparent);
            var pntStroke0 = new SvgPaint(new SvgColor(Color.Green));
            var pntStroke0inv = new SvgPaint(new SvgColor(Color.FromArgb(100, Color.Green)));
            var pntFill0 = new SvgPaint(new SvgColor(Color.FromArgb(100, Color.Orange)));
            var pt0_0 = new SvgPoint(len0, len0);
            var pt100_100 = new SvgPoint(len100, len100);

            // 円。
            var svgDoc = new GcSvgDocument();
            var svg = svgDoc.RootSvg;
            svg.Children.Add(new SvgEllipseElement()
            {
                CenterX = len50,
                CenterY = len50,
                RadiusX = len50,
                RadiusY = len50,
                Fill = paintTrans,
                Stroke = pntStroke0,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X - grid.Radius, ip.Y - grid.Radius, grid.Radius * 2, grid.Radius * 2));
            DrawCaption(g, ip, grid.Radius, "円");
            ip.X += grid.StepX;

            // 楕円。
            svg.Children.Clear();
            svg.Children.Add(new SvgEllipseElement()
            {
                CenterX = len50,
                CenterY = len50,
                RadiusX = len50,
                RadiusY = len50,
                Fill = paintTrans,
                Stroke = pntStroke0,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X - grid.Radius * 1.4f, ip.Y - grid.Radius / 2, grid.Radius * 2 * 1.4f, grid.Radius));
            DrawCaption(g, ip, grid.Radius, "楕円");
            ip.X += grid.StepX;

            // 円柱。
            svg.Children.Clear();
            float radX = grid.Radius / 1.4f;
            float radY = grid.Radius / 6;
            float height = grid.Radius * 1.8f;
            svg.Children.Add(new SvgEllipseElement()
            {
                CenterX = len50,
                CenterY = len10,
                RadiusX = len50,
                RadiusY = len10,
                Fill = paintTrans,
                Stroke = pntStroke0,
            });
            svg.Children.Add(new SvgEllipseElement()
            {
                CenterX = len50,
                CenterY = len90,
                RadiusX = len50,
                RadiusY = len10,
                Fill = pntFill0,
                Stroke = pntStroke0,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = len0,
                Y1 = len10,
                X2 = len0,
                Y2 = len90,
                Stroke = pntStroke0,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = len100,
                Y1 = len10,
                X2 = len100,
                Y2 = len90,
                Stroke = pntStroke0,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X - grid.Radius, ip.Y - grid.Radius, grid.Radius * 2, grid.Radius * 2));
            DrawCaption(g, ip, grid.Radius, "円柱");
            ip.X = startIp.X;
            ip.Y += grid.StepY;

            // 正方形。
            DrawPolygon(g, ip, grid.Radius, 4, (float)-Math.PI / 4, pntStroke0, "正方形");
            ip.X += grid.StepX;

            // 長方形。
            svg.Children.Clear();
            svg.Children.Add(new SvgRectElement()
            {
                X = new SvgLength(-20, SvgLengthUnits.Percentage),
                Y = len20,
                Width = new SvgLength(140, SvgLengthUnits.Percentage),
                Height = len60,
                Stroke = pntStroke0,
                Fill = paintTrans,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X - grid.Radius, ip.Y - grid.Radius, grid.Radius * 2, grid.Radius * 2));
            DrawCaption(g, ip, grid.Radius, "長方形");
            ip.X += grid.StepX;

            // 立方体。
            svg.Children.Clear();
            svg.Children.Add(new SvgRectElement()
            {
                X = len10,
                Y = len10,
                Width = len60,
                Height = len60,
                Stroke = pntStroke0,
                Fill = paintTrans,
            });
            svg.Children.Add(new SvgRectElement()
            {
                X = len30,
                Y = len30,
                Width = len60,
                Height = len60,
                Stroke = pntStroke0,
                Fill = paintTrans,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = len10,
                Y1 = len10,
                X2 = len30,
                Y2 = len30,
                Stroke = pntStroke0,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = len10,
                Y1 = len70,
                X2 = len30,
                Y2 = len90,
                Stroke = pntStroke0,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = len70,
                Y1 = len10,
                X2 = len90,
                Y2 = len30,
                Stroke = pntStroke0,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = len70,
                Y1 = len70,
                X2 = len90,
                Y2 = len90,
                Stroke = pntStroke0,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X - grid.Radius, ip.Y - grid.Radius, grid.Radius * 2, grid.Radius * 2));
            DrawCaption(g, ip, grid.Radius, "立方体");
            ip.X = startIp.X;
            ip.Y += grid.StepY;

            // 五角形。
            DrawPolygon(g, ip, grid.Radius, 5, (float)-Math.PI / 2, pntStroke0, "五角形");
            ip.X += grid.StepX;

            // 六角形。
            // サンプルのために、変換を適用して六角形をより幅広く短くします。
            g.Transform = Matrix3x2.CreateScale(1.4f, 0.8f) * Matrix3x2.CreateTranslation(ip.X, ip.Y);
            DrawPolygon(g, PointF.Empty, grid.Radius, 6, 0, pntStroke0, null);
            g.Transform = Matrix3x2.Identity;
            DrawCaption(g, ip, grid.Radius, "六角形");
            ip.X += grid.StepX;

            // 八角形。
            DrawPolygon(g, ip, grid.Radius, 8, (float)-Math.PI / 8, pntStroke0, "八角形");
            ip.X = startIp.X;
            ip.Y += grid.StepY;

            // 三角形。
            DrawPolygon(g, ip, grid.Radius, 3, (float)-Math.PI / 2, pntStroke0, "三角形");
            ip.X += grid.StepX;

            // 塗りつぶし五面体。
            var svgPts = DrawPolygon(g, ip, grid.Radius, 5, (float)-Math.PI / 2, pntStroke0, "五面体");
            svg.Children.Clear();
            svg.Children.Add(new SvgPolygonElement()
            {
                Points = new List<SvgPoint>()
                {
                    svgPts[0], svgPts[2], svgPts[4], svgPts[1], svgPts[3],
                },
                Stroke = pntStroke0,
                Fill = pntFill0,
                FillRule = SvgFillRule.EvenOdd,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X, ip.Y, grid.Radius * 2, grid.Radius * 2));
            ip.X += grid.StepX;

            // ピラミッドを描くために単純な斜投影を設定します。
            var angle = Math.PI / 6;
            float s = (float)Math.Sin(angle);
            float c = (float)Math.Cos(angle);
            Func<float, float, float, PointF> project = (x_, y_, z_) => new PointF(x_ - c * y_ * 0.5f, -(z_ - s * y_ * 0.5f));
            Func<Vector3, PointF> p3d = v_ => project(v_.X, v_.Y, v_.Z);
            float hedge = grid.Radius; // 1/2 edge
            // デバッグ - ３つの軸を描画します。
            /*
            g.DrawLine(project(0, 0, 0), project(100, 0, 0), Color.Red);
            g.DrawLine(project(0, 0, 0), project(0, 100, 0), Color.Green);
            g.DrawLine(project(0, 0, 0), project(0, 0, 100), Color.Blue);
            */
            // 正方形のピラミッドを形成する 3D 点です。
            var pts3d = new Vector3[]
            {
                new Vector3(-hedge, -hedge, 0),
                new Vector3(hedge, -hedge, 0),
                new Vector3(hedge, hedge, 0),
                new Vector3(-hedge, hedge, 0),
                new Vector3(0, 0, hedge * 2),
            };
            // ピラミッドを描くために点を投影します。
            var pts = pts3d.Select(v_ => p3d(v_));
            svgPts = pts.Select(p_ => new SvgPoint(new SvgLength(p_.X, SvgLengthUnits.Percentage), new SvgLength(p_.Y, SvgLengthUnits.Percentage))).ToArray();
            g.Transform = Matrix3x2.CreateTranslation(0, hedge * 0.7f);
            svg.Children.Clear();
            // 表面のエッジ。
            svg.Children.Add(new SvgPolygonElement()
            {
                Points = new List<SvgPoint>()
                {
                    svgPts[4],
                    svgPts[1],
                    svgPts[2],
                    svgPts[3],
                    svgPts[4],
                    svgPts[2],
                },
                Stroke = pntStroke0,
                Fill = paintTrans,
            });
            // 底面。
            svg.Children.Add(new SvgPolygonElement()
            {
                Points = svgPts.Take(4).ToList(),
                Fill = pntFill0,
            });
            // 影のエッジ。
            svg.Children.Add(new SvgLineElement()
            {
                X1 = svgPts[0].X,
                Y1 = svgPts[0].Y,
                X2 = svgPts[4].X,
                Y2 = svgPts[4].Y,
                Stroke = pntStroke0inv,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = svgPts[0].X,
                Y1 = svgPts[0].Y,
                X2 = svgPts[1].X,
                Y2 = svgPts[1].Y,
                Stroke = pntStroke0inv,
            });
            svg.Children.Add(new SvgLineElement()
            {
                X1 = svgPts[0].X,
                Y1 = svgPts[0].Y,
                X2 = svgPts[3].X,
                Y2 = svgPts[3].Y,
                Stroke = pntStroke0inv,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X, ip.Y, grid.Radius * 2, grid.Radius * 2));
            g.Transform = Matrix3x2.Identity;
            DrawCaption(g, ip, grid.Radius, "ピラミッド");
            ip.X = startIp.X;
            ip.Y += grid.StepY;

            // 三角錐。
            float baseh = grid.Radius * 0.3f;
            svgPts = DrawPolygon(g, ip, grid.Radius, 3, (float)-Math.PI / 2, null, "三角錐");
            svg.Children.Clear();
            svg.Children.Add(new SvgPolylineElement()
            {
                Points = new List<SvgPoint>()
                {
                    svgPts[2], svgPts[0], svgPts[1],
                },
                Stroke = pntStroke0,
                Fill = paintTrans,
            });
            svg.Children.Add(new SvgEllipseElement()
            {
                CenterX = new SvgLength(svgPts[0].X.Value, SvgLengthUnits.Percentage),
                CenterY = svgPts[2].Y,
                RadiusX = new SvgLength((svgPts[1].X.Value - svgPts[2].X.Value) / 2, SvgLengthUnits.Percentage),
                RadiusY = len10,
                Fill = pntFill0,
                Stroke = pntStroke0,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X, ip.Y, grid.Radius * 2, grid.Radius * 2));
            ip.X += grid.StepX;

            // 平行四辺形（四角形を変形して利用します）。
            svg.Children.Clear();
            var svgParal = new SvgRectElement()
            {
                X = new SvgLength(-20, SvgLengthUnits.Percentage),
                Y = len20,
                Width = new SvgLength(140, SvgLengthUnits.Percentage),
                Height = len60,
                Stroke = pntStroke0,
                Fill = paintTrans,
            };
            svg.Children.Add(svgParal);
            g.Transform = Matrix3x2.CreateSkew((float)Math.PI / 6, 0) * Matrix3x2.CreateTranslation(ip.X, ip.Y);
            g.DrawSvg(svgDoc, new RectangleF(-grid.Radius, -grid.Radius, grid.Radius * 2, grid.Radius * 2));
            g.Transform = Matrix3x2.Identity;
            DrawCaption(g, ip, grid.Radius, "平行四辺形");
            ip.X += grid.StepX;

            // 台形（DrawPolygon を使って四角形の点を取得します）。
            svg.Children.Clear();
            float dx = 10;
            svgPts = DrawPolygon(g, ip, grid.Radius, 4, (float)-Math.PI / 4, null, "台形");
            svg.Children.Add(new SvgPolygonElement()
            {
                Points = new List<SvgPoint>()
                {
                    new SvgPoint(new SvgLength(svgPts[0].X.Value - dx, SvgLengthUnits.Percentage), svgPts[0].Y),
                    new SvgPoint(new SvgLength(svgPts[1].X.Value + dx, SvgLengthUnits.Percentage), svgPts[1].Y),
                    new SvgPoint(new SvgLength(svgPts[2].X.Value - dx, SvgLengthUnits.Percentage), svgPts[2].Y),
                    new SvgPoint(new SvgLength(svgPts[3].X.Value + dx, SvgLengthUnits.Percentage), svgPts[3].Y),
                },
                Stroke = pntStroke0,
                Fill = paintTrans,
            });
            g.DrawSvg(svgDoc, new RectangleF(ip.X, ip.Y, grid.Radius * 2, grid.Radius * 2));

            // PDF ドキュメントを保存します。
            doc.Save(stream);
            return doc.Pages.Count;
        }
    }
}
